from typing import List, Dict, Optional

from kama_sdk.core.core import job_client, utils, consts
from kama_sdk.core.core.types import CommitOutcome, PredEval
from kama_sdk.model.action.base.action import Action
from kama_sdk.model.base.model import Model
from kama_sdk.model.operation.field import Field
from kama_sdk.model.operation.operation_state import OperationState
from kama_sdk.model.operation.step_state import StepState

TOS = OperationState
TSS = StepState
TOOS = Optional[OperationState]
TCO = CommitOutcome


class Step(Model):

  def sig(self) -> str:
    parent_id = self.parent.id() if self.parent else 'orphan'
    return f"{parent_id}::{self.id()}"

  def runs_action(self) -> bool:
    return bool(self.config.get(ACTION_KEY))

  def next_step_id(self, op_state: TOS) -> Optional[str]:
    self.patch(gen_downstream_patches(op_state, {}))
    return self.get_prop(NEXT_STEP_KEY)

  def validate_field(self, field_id: str, value: str, op_state: TOS) -> PredEval:
    fields = self.inflate_children(
      Field,
      prop=FIELDS_KEY,
      patches=gen_downstream_patches(op_state, {field_id: value})
    )
    finder = lambda f: f.id() == field_id
    field = next(filter(finder, fields), None)
    return field.variable.validate(value)

  def visible_fields(self, assignments, op_state: TOS) -> List[Field]:
    fields = self.inflate_children(
      Field,
      prop=FIELDS_KEY,
      patches=gen_downstream_patches(op_state, assignments)
    )
    return [f for f in fields if f.compute_visibility()]

  def run(self, assigns: Dict, state: StepState) -> Optional[str]:
    buckets = self.partition_vars(assigns, state.parent_op)
    state.notify_vars_assigned(buckets)

    if action := self.gen_final_action(state.parent_op, buckets):
      return job_client.enqueue_action(action.serialize())
    else:
      return None

  def gen_final_action(self, op_state: OperationState, buckets) -> Optional[Action]:
    """
    In case the action has dynamic parts to it , we must resolve it
    before it runs in a different process (no shared memory). Also merges in
    event-telem config. Omits the unmarshalable 'context' of the config.
    @param op_state: step state that IFTT might use to make resolution
    @param buckets: value buckets eg chart, inline, state
    @return: un-IFTT'ed config dict for the action
    """
    return self.inflate_child(
      Action,
      prop=ACTION_KEY,
      safely=True,
      patches=dict(
        **gen_downstream_patches(op_state, {}),
        commit={
          consts.TARGET_STANDARD: buckets[consts.TARGET_STANDARD],
          consts.TARGET_INLIN: buckets[consts.TARGET_INLIN],
          consts.TARGET_PREFS: buckets[consts.TARGET_PREFS]
        }
      )
    ) if self.runs_action() else None

  def partition_vars(self, inputs: Dict, op_state: TOS) -> Dict:
    self.patch(gen_downstream_patches(op_state, inputs))
    default_commit = self._partition_vars_by_field(inputs, op_state)
    self.patch(dict(commit=default_commit))
    commit_override = self.resolve_prop('commit', depth=100)
    return sanitize_commit_bundle(commit_override, default_commit)

  def _partition_vars_by_field(self, inputs: Dict, op_state: TOS) -> Dict:
    fields = self.visible_fields(inputs, op_state)
    types = consts.TARGET_TYPES
    return {t: filter_inputs_by_type(fields, inputs, t) for t in types}


def gen_downstream_patches(op_state: OperationState, user_input: Dict):
  return dict(
    op_state=op_state.all_assigns(),
    inputs=user_input
  )


def sanitize_commit_bundle(overrides: Dict, defaults: Dict) -> Dict:
  output = {}
  for key, default in defaults.items():
    override = (overrides or {}).get(key)
    if override == {}:
      final_dict = {}
    elif override is None:
      final_dict = default
    else:
      final_dict = override
    output[key] = utils.flat2deep(final_dict)

  for mandatory_type in consts.TARGET_TYPES:
    if mandatory_type not in output.keys():
      output[mandatory_type] = {}

  return output


def filter_inputs_by_type(fields: List[Field], inputs: Dict, _type: str):
  flat_inputs = utils.deep2flat(inputs)

  def find_field(_id):
    return next(filter(lambda f: f.id() == _id, fields), None)

  gate = lambda k: find_field(k) and find_field(k).target == _type
  filtered_flat = {k: v for (k, v) in flat_inputs.items() if gate(k)}
  return utils.flat2deep(filtered_flat)


ACTION_KEY = 'action'
NEXT_STEP_KEY = 'next'
FIELDS_KEY = 'fields'



from abc import ABC

from werkzeug.utils import cached_property

from kama_sdk.model.base.model import Model


class QuantityHumanizer(Model, ABC):

  ROUNDING_KEY = 'rounding'
  PREFIX_KEY = 'prefix'
  SUFFIX_KEY = 'suffix'

  @cached_property
  def rounding(self):
    return self.get_prop(self.ROUNDING_KEY, 0)

  @cached_property
  def prefix(self):
    return self.get_prop(self.PREFIX_KEY, '')

  @cached_property
  def suffix(self):
    return self.get_prop(self.SUFFIX_KEY, '')

  def humanize_quantity(self, raw_quantity: float) -> float:
    better_quantity = self._humanize_quantity(raw_quantity)
    if self.rounding > 0:
      return round(better_quantity, self.rounding)
    else:
      return int(better_quantity)

  def humanize_expr(self, raw_value: float) -> str:
    better_expr = self._humanize_expr(raw_value)
    return f"{self.prefix}{better_expr}{self.suffix}"

  def _humanize_expr(self, raw_value: float) -> str:
    return str(self.humanize_quantity(raw_value))

  @staticmethod
  def _humanize_quantity(value: float) -> float:
    return value



from typing import List

from kama_sdk.core.core.types import KAO
from kama_sdk.model.action.ext.manifest.await_predicates_settled_action import AwaitPredicatesSettledAction
from kama_sdk.model.supplier.predicate.outkomes2preds import outkomes2preds
from kama_sdk.model.supplier.predicate.predicate import Predicate


class AwaitOutkomesSettledAction(AwaitPredicatesSettledAction):

  def outkomes(self) -> List[KAO]:
    return self.get_prop('outkomes')

  def predicates(self) -> List[Predicate]:
    return outkomes2preds(self.outkomes())



from typing import List

from werkzeug.utils import cached_property

from kama_sdk.core.core.types import Reconstructor
from kama_sdk.model.concern.concern_view_adapter import ConcernViewAdapter
from kama_sdk.model.concern.concern_card_adapter import ConcernCardAdapter
from kama_sdk.model.concern.concern_set import ConcernSet


class ConcernGrid(ConcernSet):

  def compute_partial(self):
    pass

  def compute_concern_seeds(self):
    return

  def compute_single(self, tuple_id: str):
    pass

  @cached_property
  def reconstructors(self) -> List[Reconstructor]:
    return self.get_prop(CARDS_KEY, [])

  def one_shot_compute(self):
    output = []
    for reconstructor in self.reconstructors:
      adapter = ConcernCardAdapter.reconstruct(reconstructor)
      output.append(adapter.compute())
    return output


  def compute(self):
    return {}

  @classmethod
  def view_type(cls) -> str:
    return 'grid'


CARDS_KEY = 'cards'


from typing import List

from werkzeug.utils import cached_property

from kama_sdk.core.core.types import KoD, ErrCapture
from kama_sdk.core.core.utils import any2bool
from kama_sdk.model.action.base.action import Action, ActionError
from kama_sdk.model.action.base.multi_action import MultiAction
from kama_sdk.model.supplier.predicate.predicate import Predicate

class RunPredicateAction(Action):

  @classmethod
  def global_provider_id(cls):
    return 'sdk.run-predicates-action.master-provider'

  def id(self) -> str:
    from_super = super().id()
    return from_super or self.predicate().id()

  @cached_property
  def abort_on_fail(self) -> bool:
    raw_value = self.get_prop('abort_on_fail', False)
    return any2bool(raw_value)

  def predicate(self) -> Predicate:
    return self.inflate_child(
      Predicate,
      prop='predicate',
      resolve_kod=False
    )

  @cached_property
  def title(self):
    return super().title or self.predicate().title

  @cached_property
  def info(self):
    return super().info or self.predicate().info

  def perform(self) -> None:
    predicate = self.predicate()
    if not any2bool(predicate.resolve()):
      raise ActionError(ErrCapture(
        fatal=self.abort_on_fail,
        type='negative_predicate',
        name=predicate.id(),
        reason=predicate.reason(),
        extras=dict(
          predicate_id=predicate.id(),
          predicate_kind=predicate.kind(),
          predicate_challlenge=predicate.challenge(),
          predicate_check_against=predicate.check_against(),
          **predicate.error_extras()
        )
      ))


class RunPredicatesAction(MultiAction):

  def predicates(self) -> List[Predicate]:
    return self.inflate_children(
      Predicate,
      prop='predicates',
      resolve_kod=False
    )

  @cached_property
  def sub_actions(self) -> List[Action]:
    kods = [predicate.serialize() for predicate in self.predicates()]
    return [self.pred_kod2action(kod) for kod in kods]

  def pred_kod2action(self, predicate_kod: KoD) -> RunPredicateAction:
    action = RunPredicateAction(dict(predicate=predicate_kod))
    action.parent = self
    return action



mock_app_update_id = "sdk.provider.mock.ktea-update"
mock_injection_bundle_id = "sdk.provider.mock.injection"
mock_initial_injection_bundle_id = "sdk.provider.mock.initial-injection"
deletion_spec_id = 'sdk.provider.deletion-spec'



from typing import List, Dict, Optional

from typing_extensions import TypedDict

from kama_sdk.core.core import utils
from kama_sdk.core.core.types import KAO, ErrCapture, K8sResDict
from kama_sdk.core.ktea.ktea_client import KteaClient
from kama_sdk.model.action.base.action import Action
from kama_sdk.model.action.base.action_errors import FatalActionError


class OutputFormat(TypedDict):
  outkomes: List[KAO]


ResDescs = List[K8sResDict]

class KubectlApplyAction(Action):

  def res_descs(self) -> List[Dict]:
    return self.get_prop('res_descs')

  def perform(self) -> OutputFormat:
    res_descs = self.res_descs()
    outkomes = KteaClient.kubectl_apply(res_descs)
    self.add_logs(list(map(utils.kao2log, outkomes)))
    check_kao_failures(outkomes, res_descs)
    return dict(outkomes=outkomes)


def is_outkome_error(outkome: KAO) -> bool:
  return outkome.get('error') is not None


def culprit_res_desc(culprit: KAO, descs: ResDescs) -> Optional[K8sResDict]:
  matcher = lambda desc: utils.same_res(culprit, utils.full_res2sig(desc))
  return next(filter(matcher, descs), None)


def check_kao_failures(outcomes: List[KAO], res_descs: ResDescs):
  if kao_culprit := next(filter(is_outkome_error, outcomes), None):
    res_name = kao_culprit.get('name')
    res_kind = kao_culprit.get('kind')

    raise FatalActionError(ErrCapture(
      type='kubectl_apply',
      name='resource_rejected',
      reason=f'kubectl apply failed for {res_kind}/{res_name}',
      logs=[kao_culprit.get('error')],
      extras=dict(
        resource_signature=dict(name=res_name, kind=res_kind),
        resource=culprit_res_desc(kao_culprit, res_descs)
      )
    ))



from typing import Dict

from werkzeug.utils import cached_property

from kama_sdk.core.core import config_man as cm, utils
from kama_sdk.core.core.config_man import config_man
from kama_sdk.core.core.types import ErrCapture
from kama_sdk.model.action.base.action import Action
from kama_sdk.model.action.base.action_errors import FatalActionError


class PatchManifestVarsAction(Action):

  @cached_property
  def title(self) -> str:
    return super().title or f"Patch {self.source_key} values"

  @cached_property
  def info(self) -> str:
    return super().title or f"Merge new value into existing {self.source_key}"

  def values(self) -> Dict:
    return utils.flat2deep(self.get_prop('values') or {})

  def source_key(self) -> str:
    return self.get_prop('target_key', cm.user_vars_key)

  def perform(self) -> None:
    config_man.patch_into_deep_dict(
      self.source_key(),
      self.values()
    )


class WriteManifestVarsAction(Action):

  def values(self) -> Dict:
    return self.get_prop('values') or {}

  def source_key(self) -> str:
    return self.get_prop('target_key')

  def perform(self) -> None:
    self.raise_if_illegal_source_key()
    config_man.patch_master_cmap(
      self.source_key(),
      self.values()
    )

  def raise_if_illegal_source_key(self):
    reason = None
    source_key = self.source_key()
    if not source_key:
      reason = "source key must be explicit for dangerous action"
    elif source_key not in cm.manifest_var_keys:
      reason = f"source key {source_key} not in {cm.manifest_var_keys}"

    if reason:
      raise FatalActionError(ErrCapture(
        type='write_manifest_action_illegal_key',
        reason=reason
      ))



from werkzeug.utils import cached_property

from kama_sdk.model.glance.glance import Glance


class GraphicAndTextGlance(Glance):

  @cached_property
  def view_type(self):
    return 'resource'

  @cached_property
  def line_one(self):
    return self.get_prop(LINE_ONE_KEY)

  @cached_property
  def line_two(self):
    return self.get_prop(LINE_TWO_KEY)

  @cached_property
  def line_three(self):
    return self.get_prop(LINE_THREE_KEY)

  @cached_property
  def graphic(self):
    return self.get_prop(GRAPHIC_KEY)

  @cached_property
  def graphic_type(self):
    return self.get_prop(GRAPHIC_TYPE_KEY, 'icon')

  @cached_property
  def graphic_emotion(self):
    return self.get_prop(GRAPHIC_EMOTION_KEY, 'primaryColor')

  @cached_property
  def small_graphic(self) -> bool:
    return self.get_prop(SMALL_GRAPHIC_KEY, False)

  def content_spec(self):
    return {
      'graphic_type': self.graphic_type,
      'graphic': self.graphic,
      'line_one': self.line_one,
      'small': self.small_graphic,
      'line_two': self.line_two,
      'line_three': self.line_three,
      'graphic_emotion': self.graphic_emotion
    }


LINE_ONE_KEY = 'line_one'
LINE_TWO_KEY = 'line_two'
LINE_THREE_KEY = 'line_three'
GRAPHIC_KEY = 'graphic'
GRAPHIC_TYPE_KEY = 'graphic_type'
GRAPHIC_EMOTION_KEY = 'graphic_emotion'
SMALL_GRAPHIC_KEY = 'small_graphic'



import json
import os
from datetime import datetime
from pathlib import Path
from typing import Optional, Dict, Type, Any

from typing_extensions import TypedDict

from kama_sdk.core.core import utils

class CmapAccessRecord(TypedDict):
  read_ts: datetime
  write_ts: datetime


def serialize_outbound(value: Any, expected_type: Type) -> str:
  if expected_type == bool:
    return str(utils.any2bool(value))
  elif expected_type == str:
    return value or ''
  elif expected_type == dict:
    return json.dumps(value)
  elif expected_type == datetime:
    return value.strftime(iso8601_time_fmt)
  else:
    print(f"[kama_sdk:patch] bad type {expected_type} for {value}")
    return str(value)

def parse_inbound(raw_value: str, type_mapping: Type) -> Any:
  if type_mapping == dict:
    return parse_dict(raw_value)
  elif type_mapping == bool:
    return utils.any2bool(raw_value)
  elif type_mapping == datetime:
    return parse_ts(raw_value)
  else:
    return raw_value

def parse_ts(raw_value, backup=None) -> Optional[datetime]:
  parsed = None
  if raw_value:
    try:
      parsed = datetime.strptime(raw_value, iso8601_time_fmt)
    except TypeError:
      pass
  return parsed or backup


def parse_dict(raw_value) -> Dict:
  backup = {}
  try:
    return json.loads(raw_value or '{}') or backup
  except:
    print(f"[kama_sdk] DANGER corrupt dict {raw_value}")
    return backup


def does_type_match(value: Any, expected: Type) -> bool:
  if expected == bool:
    strings = ['true', 'false', 'True', 'False']
    return value in [True, False, None,  *strings]
  else:
    return type(value) == expected


def track_cmap_read(ns: str):
  update_tracker(ns, 'read_ts')


def track_cmap_write(ns: str):
  update_tracker(ns, 'write_ts')


def update_tracker(ns: str, key: str):
  crt = read_tracker(ns)
  updated_tracker = {**crt, 'ns': ns, key: datetime.now()}
  with open(io_tracker_fname, 'w') as file:
    serialized = serialize_tracker(updated_tracker)
    file.write(json.dumps(serialized))


def serialize_tracker(tracker: CmapAccessRecord) -> Dict:
  return {
    **tracker,
    'read_ts': tracker['read_ts'].strftime(iso8601_time_fmt),
    'write_ts': tracker['write_ts'].strftime(iso8601_time_fmt)
  }


def is_cmap_dirty(ns: str):
  tracker = read_tracker(ns)
  return tracker['write_ts'] > tracker['read_ts']


def ancient_dt(offset=0) -> datetime:
  date_time_str = f'2000-01-01 0{offset}:00:00.000000'
  return datetime.strptime(date_time_str, iso8601_time_fmt)


def read_tracker(ns: str) -> CmapAccessRecord:
  Path(io_tracker_fname).touch(exist_ok=True)
  with open(io_tracker_fname, 'r+') as file:
    contents = file.read() or '{}'
    try:
      parsed = json.loads(contents)
      if not parsed.get('ns') == ns:
        parsed = {}
    except:
      parsed = {}
    return {
      'read_ts': parse_ts(parsed.get('read_ts'), ancient_dt(0)),
      'write_ts': parse_ts(parsed.get('write_ts'), ancient_dt(1))
    }


def clear_trackers():
  if os.path.exists(io_tracker_fname):
    try:
      os.remove(io_tracker_fname)
    except FileNotFoundError:
      msg = f"race condition on {io_tracker_fname}"
      print(f"[kama_sdk:config_man] WARN {msg}, process restart recommended")


iso8601_time_fmt = '%Y-%m-%d %H:%M:%S.%f'
io_tracker_fname = '/tmp/kama_io_tracker.json'



from typing import Optional

from kama_sdk.core.core import updates_man
from kama_sdk.core.core.types import InjectionsDesc
from kama_sdk.model.supplier.base.supplier import Supplier


class LatestVendorInjectionSupplier(Supplier):
  def _compute(self) -> Optional[InjectionsDesc]:
    return updates_man.latest_injection_bundle()



import json
from copy import deepcopy
from typing import Optional, Dict

from bson import ObjectId
from pymongo import MongoClient
from pymongo.database import Database
from pymongo.errors import ServerSelectionTimeoutError
from pymongo.results import InsertOneResult

from kama_sdk.core.core import hub_api_client, utils
from kama_sdk.core.core.config_man import config_man
from kama_sdk.core.core.types import ErrCapture, EventCapture


def database() -> Optional[Database]:
  return connection_obj[key_conn_obj_db]


def create_session_if_none() -> Optional[Database]:
  if not database():
    prefs = config_man.prefs()
    strategy = utils.deep_get2(prefs, key_telem_strategy)
    is_enabled = not strategy == strategy_disabled
    if utils.is_out_of_cluster() or is_enabled:
      connection_obj[key_conn_obj_db] = connect()
  return database()


def clear_session():
  del connection_obj[key_conn_obj_db]


def is_storage_ready() -> bool:
  return True if create_session_if_none() else False


def parametrized(dec):
  def layer(*args, **kwargs):
    def repl(f):
      return dec(f, *args, **kwargs)
    return repl
  return layer


@parametrized
def connected_and_enabled(func, backup=None):
  def aux(*xs, **kws):
    return func(*xs, **kws) if create_session_if_none() else backup
  return aux


def store_error(error: ErrCapture) -> InsertOneResult:
  clean_error = {
    **deepcopy(error),
    'extras': json.dumps(error.get('extras'))
  }
  clean_error.pop('is_original', None)
  return store_list_element(ERRORS_COLLECTION_NAME, clean_error)


def store_event(event: EventCapture) -> InsertOneResult:
  return store_list_element(EVENTS_COLLECTION_NAME, event)


def store_config_backup(outcome: Dict) -> InsertOneResult:
  return store_list_element(BACKUPS_COLLECTION_KEY, outcome)


def list_errors():
  return list_records(ERRORS_COLLECTION_NAME)


def list_events():
  return list_records(EVENTS_COLLECTION_NAME)


def list_config_backups():
  return list_records(BACKUPS_COLLECTION_KEY)


def find_event(query: Dict) -> Optional[EventCapture]:
  return find_record(EVENTS_COLLECTION_NAME, query)


def find_error(query: Dict) -> Optional[ErrCapture]:
  return find_record(ERRORS_COLLECTION_NAME, query)


def get_config_backup(record_id) -> Optional[Dict]:
  return find_record_by_id(BACKUPS_COLLECTION_KEY, record_id)


@connected_and_enabled(backup=None)
def store_list_element(list_key: str, item: Dict) -> InsertOneResult:
  item = {**item, key_synced: False}
  return database()[list_key].insert_one(item)


@connected_and_enabled(backup=[])
def list_records(key: str):
  return list(database()[key].find())


@connected_and_enabled(backup=[])
def clear_update_outcomes():
  database().delete(EVENTS_COLLECTION_NAME)


@connected_and_enabled(backup=None)
def find_record(collection_id: str, query: Dict):
  collection = database()[collection_id]
  return collection.find_one(query)


@connected_and_enabled(backup=None)
def find_record_by_id(collection_id, record_id):
  return find_record(collection_id, {'_id': ObjectId(record_id)})


def upload_all_meta():
  upload_status()
  if is_storage_ready():
    upload_events_and_errors()
  else:
    print("[kama_sdk:telem_man] db unavailable, skip upload events/errors")


def upload_status() -> bool:
  if config_man.is_training_mode():
    return False

  ktea = config_man.ktea_desc()
  kama = config_man.kama_desc()

  payload = {
    'status': config_man.application_status(),
    'ktea_type': ktea.get('type'),
    'ktea_version': ktea.get('version'),
    'kama_type': kama.get('type'),
    'kama_version': kama.get('version'),
    'synced_at': str(config_man.last_updated()),
  }

  print(f"[kama_sdk:telem_man] stat -> {hub_api_client.backend_host()} {payload}")
  response = hub_api_client.patch('/install', dict(install=payload))
  print(f"[kama_sdk:telem_man] upload status resp {response}")
  return response.status_code < 205


def upload_events_and_errors():
  if config_man.is_training_mode():
    return False

  if utils.is_test():
    return False

  session = create_session_if_none()

  if not session:
    return False

  for collection_name in [ERRORS_COLLECTION_NAME, EVENTS_COLLECTION_NAME]:
    collection = session[collection_name]
    items = collection.find({key_synced: False})
    for item in items:
      hub_key = f'kama_{collection_name}'[0:-1]
      clean_item = deepcopy(item)
      clean_item.pop('_id', None)
      clean_item.pop('synced', None)
      resp = hub_api_client.post(f'/{hub_key}s', {hub_key: clean_item})
      if not resp.status_code == 400:
        query = {'_id': item['_id']}
        patch = {'$set': {key_synced: True}}
        collection.update_one(query, patch)


def connect() -> Optional[Database]:
  if utils.is_in_cluster():
    manifest_vars = config_man.user_vars()
    host = manifest_vars.get('telem_db.host', 'telem-db')
    port = manifest_vars.get('telem_db.port', 27017)
  else:
    host = 'localhost'
    port = 27017

  try:
    client = MongoClient(
      host=host,
      port=int(port),
      connectTimeoutMS=1_000,
      serverSelectionTimeoutMS=1_000
    )
    client.server_info()
    return client[f'{utils.run_env()}_database']
  except ServerSelectionTimeoutError:
    if not utils.is_test():
      # print(f"[kama_sdk:telem_man] MongoDB[{host}, {port}] conn failed")
      if utils.is_out_of_cluster():
        pass
        # print("For local dev server, run MongoDB on localhost:27017")
      return None


key_conn_obj_db = 'db'
key_synced = 'synced'

EVENTS_COLLECTION_NAME = 'events'
BACKUPS_COLLECTION_KEY = 'config_backups'
ERRORS_COLLECTION_NAME = 'errors'

key_telem_strategy = 'telem_db.strategy'
strategy_disabled = 'disabled'
strategy_internal = 'managed_pvc'

connection_obj = {key_conn_obj_db: None}



from kama_sdk.model.operation.field import Field

def document():
  Field.is_manifest_bound.__doc__ = """
  Checks if the variable should be recorded in the manifest.
  :return: True if it should, False otherwise.
  """



from typing import List, Any, Dict

from kama_sdk.core.core.utils import pwar
from kama_sdk.model.action.base.action import Action
from kama_sdk.model.base.model import Model
from kama_sdk.model.concern.concern_set import ConcernSet
from kama_sdk.model.concern.concern_view_adapter import ConcernViewAdapter
from kama_sdk.model.concern.concern_attr_adapter import ConcernAttrAdapter
from kama_sdk.model.operation.operation import Operation


class AttrPanelDescriptor(Model):

  def attr_adapters(self) -> List[ConcernAttrAdapter]:
    return self.inflate_children(
      ConcernAttrAdapter,
      prop=ATTRS_ADAPTERS_KEY,
    )

  def compute_values(self) -> Dict[str, Any]:
    return {a.label: a.compute_value() for a in self.attr_adapters()}


class ConcernDetailAdapter(ConcernViewAdapter):

  def attr_panel_descs(self) -> List[AttrPanelDescriptor]:
    return self.inflate_children(
      AttrPanelDescriptor,
      prop=ATTR_PANEL_DESCRIPTORS_KEY
    )

  def value_panel_descs(self) -> List[ConcernAttrAdapter]:
    return self.inflate_children(
      ConcernAttrAdapter,
      prop=VALUE_PANEL_DESCRIPTORS_KEY
    )

  def related_sets(self) -> List[ConcernSet]:
    return self.inflate_children(
      ConcernSet,
      prop=RELATED_SETS_KEY
    )

  def operations(self) -> List[Operation]:
    return self.inflate_children(
      Operation,
      prop=OPERATIONS_KEY
    )

  def actions(self) -> List[Action]:
    return self.inflate_children(
      Action,
      prop=OPERATIONS_KEY
    )

  def compute_group_data(self, group_type, group_id):
    if group_type == 'attr_panels':
      return self.compute_attr_panel_data(group_id)
    elif group_type == 'value_panels':
      return self.compute_value_panel_data(group_id)
    elif group_type == 'related_sets':
      return self.compute_related_set_data(group_id)

  def compute_attr_panel_data(self, group_id: str):
    finder = lambda p: p.id() == group_id
    if panel_model := next(filter(finder, self.attr_panel_descs())):
      return panel_model.compute_values()
    else:
      pwar(self, f"group {group_id} not found")
      return None

  def compute_value_panel_data(self, group_id: str):
    finder = lambda p: p.id() == group_id
    if panel_model := next(filter(finder, self.value_panel_descs())):
      return panel_model.compute_value()
    else:
      pwar(self, f"group {group_id} not found")
      return None

  def compute_related_set_data(self, group_id: str):
    finder = lambda p: p.id() == group_id
    if model := next(filter(finder, self.related_sets())):
      return model.one_shot_compute()
    else:
      pwar(self, f"group {group_id} not found")
      return None


ATTR_PANEL_DESCRIPTORS_KEY = 'attribute_panels'
VALUE_PANEL_DESCRIPTORS_KEY = 'value_panels'
ATTRS_ADAPTERS_KEY = 'attributes'
OPERATIONS_KEY = 'operations'
ACTIONS_KEY = 'actions'
RELATED_SETS_KEY = 'related_sets'



from kama_sdk.model.operation import predicate_statuses_computer

predicate_statuses_computer.eval_pred.__doc__ = """
  Evaluates the passed condition and prepares an output with details of the
  evaluation.
  :param predicate: condition to be evaluated.
  :return: dict with results of evaluation.
"""

predicate_statuses_computer.all_conditions_met.__doc__ = """
Checks that all conditions in the passed list are met.
:param conditions: list of conditions statuses, as TECS instances.
:return: True if all conditions are True, else False.
"""

predicate_statuses_computer.any_condition_met.__doc__ = """
  Checks that at least one condition in the passed list is met.
  :param conditions: list of conditions statuses, as TECS instances.
  :return: True if at lease one condition is True, else False.
  """



from typing import List, Optional

from k8kat.res.base.kat_res import KatRes
from werkzeug.utils import cached_property

from kama_sdk.core.core import utils
from kama_sdk.core.core.types import ErrCapture
from kama_sdk.model.action.base.action import Action, ActionError
from kama_sdk.model.action.base.multi_action import MultiAction
from kama_sdk.model.action.ext.misc.wait_action import WaitAction
from kama_sdk.model.supplier.ext.biz.resource_selector import ResourceSelector


class DeleteResourceAction(Action):

  @cached_property
  def abort_on_fail(self) -> bool:
    return bool(self.get_prop('abort_on_fail', False))

  @cached_property
  def wait_until_gone(self) -> bool:
    return bool(self.get_prop('wait_until_gone', False))

  @cached_property
  def title(self) -> str:
    if self.kat_res:
      res = self.kat_res
      return f"Delete {res.kind}/{res.name}"
    else:
      return "Victim resource does not exist"

  @cached_property
  def info(self) -> str:
    verb = "" if self.wait_until_gone else "do not"
    return f"Delete the resource, {verb} wait until it is destroyed"

  @cached_property
  def selector(self) -> ResourceSelector:
    return self.inflate_child(ResourceSelector, prop='selector')

  @cached_property
  def kat_res(self) -> Optional[KatRes]:
    _kat_res = self.config.get('kat_res')
    if not _kat_res:
      q_results = self.selector.query_cluster()
      _kat_res = q_results[0] if len(q_results) > 0 else None
    return _kat_res

  def perform(self) -> None:
    victim = self.kat_res
    if victim:
      victim.delete(self.wait_until_gone)
    else:
      raise ActionError(ErrCapture(
        fatal=self.abort_on_fail,
        type='delete_event_no_res'
      ))


class DeleteResourcesAction(MultiAction):

  def selectors(self) -> List[ResourceSelector]:
    return self.inflate_children(ResourceSelector, prop='selectors')

  @cached_property
  def child_wait_action(self) -> Optional[Action]:
    kod = self.get_prop('wait_action', default_wait_action_config)
    return self.inflate_child(WaitAction, kod=kod) if kod else None

  @cached_property
  def sub_actions(self) -> List[Action]:
    resources = self.selected_resources()
    deletion_actions = [self.kat_rest2action(r) for r in resources]
    if self.child_wait_action:
      return [*deletion_actions, self.child_wait_action]
    else:
      return deletion_actions

  def selected_resources(self) -> List[KatRes]:
    results = [s.query_cluster() for s in self.selectors()]
    return utils.flatten(results)

  def kat_rest2action(self, kat_res: KatRes) -> DeleteResourceAction:
    config = dict(
      id=f"{self.id()}.{kat_res.kind}.{kat_res.name}",
      kat_res=kat_res
    )
    return self.inflate_child(DeleteResourceAction, kod=config)


default_wait_action_config = dict(
  duration_seconds=3
)



from datetime import datetime
from typing import Dict, List, Optional

from kama_sdk.core.core.types import PromMatrixEntry
from kama_sdk.model.supplier.base.supplier import Supplier


class PromMatrixToSeriesSupplier(Supplier):

  def source_data(self) -> List[PromMatrixEntry]:
    return super(PromMatrixToSeriesSupplier, self).source_data()

  def _compute(self) -> Optional[List]:
    if result := self.source_data():
      as_time_series = matrix2series(result)
      return as_time_series
    else:
      return None

def matrix2series(matrix_entries: List[PromMatrixEntry]) -> List:
  output = []
  # warn_agg_series(matrix_entries)
  for matrix_entry in matrix_entries:
    key = infer_series_key(matrix_entry['metric'])
    for datapoint in matrix_entry['values']:
      if len(datapoint) == 2:
        epoch, computed_val = datapoint
        entry = find_or_create_entry(output, epoch)
        entry[key] = float(computed_val)
      else:
        print(f"[kama_sdk:prom_series_computer] !=2 entry val {datapoint}")

  for datapoint in output:
    epoch = datapoint['epoch']
    del datapoint['epoch']
    datapoint['timestamp'] = str(datetime.fromtimestamp(epoch))

  return output


def infer_series_key(metric: Dict) -> str:
  as_items = list(metric.items())
  if len(as_items) == 1:
    return as_items[0][1]
  elif len(as_items) == 0:
    return 'value'
  else:
    print(f"[kama_sdk:prom_series_computer] can't handle n-dim metric {metric}")
    return 'value'


def subserie_for_metric(subseries: List, metric_key: str) -> List:
  if metric_key == 'value':
    if len(subseries) == 1:
      return subseries[0]
    else:
      print("DANGER key is 'value' but +1 metric types!")
      return []

  for sub_series in subseries:
    if len(sub_series['metric']) > 0:
      if sub_series['metric'].values()[0] == metric_key:
        return sub_series
    elif len(sub_series['metric']) == 0:
      return sub_series

  print(f"DANGER no subseries for {metric_key}")
  return []


# def warn_agg_series(query_result: List[Dict]):
#   pass
  # metric_key_sets = [set(g['metric'].keys()) for g in query_result]
  # if len(metric_key_sets) >= 2:
  #   for i in range(len(metric_key_sets)):
  #     pass


def find_or_create_entry(output: List, epoch: int) -> Dict:
  for datapoint in output:
    if datapoint['epoch'] == epoch:
      return datapoint

  datapoint = {'epoch': epoch}
  output.append(datapoint)
  return datapoint



import json
import time
import traceback
from datetime import datetime, timedelta
from json import JSONDecodeError
from typing import Optional, Dict, Tuple
from urllib.parse import quote

import requests
from k8kat.res.svc.kat_svc import KatSvc

from kama_sdk.core.core.config_man import config_man
from kama_sdk.core.core.types import PromData


def compute_vector(*args) -> Optional[PromData]:
  path, args = instant_path_and_args(*args)
  return do_invoke(path, args)


def compute_matrix(*args) -> Optional[PromData]:
  path, args = gen_series_args(*args)
  # print("PATH")
  # print(path)
  # print(args)
  return do_invoke(path, args)


def do_invoke(path: str, url_args: Dict) -> Optional[PromData]:
  if is_in_cluster():
    if svc := find_prom_svc():
      if is_proxy():
        response = invoke_svc(svc, path, url_args)
      else:
        response = invoke_url(path, url_args)
    else:
      return None
  else:
    response = invoke_url(path, url_args)

  if response:
    return response.get('data')


def invoke_url(path, args: Dict) -> Optional[Dict]:
  base_url = prom_config().get(URL_KEY)
  full_url = f"{base_url}{path}?{dict_args2str(args)}"
  resp = requests.get(full_url)
  if resp.ok:
    try:
      return resp.json()
    except JSONDecodeError:
      print(traceback.format_exc())
      print(resp)
      print(f"[kama_sdk:prom_client] svc resp decode ^^ fail")
      return None


def dict_args2str(args: Dict) -> str:
  as_list = [f"{k}={quote(str(v))}" for k, v in list(args.items())]
  return "&".join(as_list)


def invoke_svc(svc: KatSvc, path: str, url_args: Dict) -> Optional[Dict]:
  resp = svc.proxy_get(path, url_args) or {}
  if resp.get('status', 500) < 300:
    try:
      return json.loads(resp.get('body'))
    except JSONDecodeError:
      print(traceback.format_exc())
      print(resp.get('body'))
      print(f"[kama_sdk:prom_client] svc resp decode ^^ fail")
      return None


def invoke_pure_http(path, args) -> Optional[Dict]:
  arg2s = lambda arg: f"{arg[0]}={arg[1]}"
  url = f"{path}?{'&'.join(list(map(arg2s, args.items())))}"
  # noinspection PyBroadException
  try:
    return requests.get(url).json()
  except:
    print(traceback.format_exc())
    print(f"[kama_sdk:prom_client] pure http invoke ^^ fail")
    return None


def instant_path_and_args(query: str, ts: datetime = None) -> Tuple:
  base = '/api/v1/query'
  ts = ts or datetime.now()
  args = {
    'query': query,
    'time': fmt_time(ts)
  }
  return base, args


def gen_series_args(query: str, step=None, t0=None, tn=None) -> Tuple:
  base = '/api/v1/query_range'
  t_start = t0 or datetime.now() - timedelta(days=7)
  t_end = tn or datetime.now()
  args = {
    'query': query,
    'start': fmt_time(t_start),
    'end': fmt_time(t_end),
    'step': step or '1h'
  }
  return base, args


def fmt_time(timestamp: datetime):
  return int(time.mktime(timestamp.timetuple()))


def find_prom_svc() -> Optional[KatSvc]:
  if not _cache_obj[CACHE_SVC_KEY]:
    prefs = prom_config()
    prom_ns = prefs.get(SVC_NS_KEY)
    prom_name = prefs.get(SVC_NAME_KEY)
    _cache_obj[CACHE_SVC_KEY] = KatSvc.find(prom_name, prom_ns)
    if not _cache_obj['svc']:
      print(f"[kama_sdk:prom_client] svc [{prefs}] not found")
  return _cache_obj[CACHE_SVC_KEY]


def prom_config() -> Dict:
  if not _cache_obj[CACHE_CONFIG_KEY]:
    root = config_man.prefs().get('monitoring') or {}
    _cache_obj[CACHE_CONFIG_KEY] = root
  return _cache_obj[CACHE_CONFIG_KEY] or {}


def is_proxy() -> bool:
  return prom_config().get(IS_PROXY_KEY, False)

def is_in_cluster() -> bool:
  return prom_config().get(LOCATION_KEY, 'out') == 'in'


URL_KEY = 'url'
IS_PROXY_KEY = 'proxy'
SVC_NS_KEY = 'service_namespace'
SVC_NAME_KEY = 'service_name'

CACHE_CONFIG_KEY = 'prom_config'
CACHE_SVC_KEY = 'svc'
LOCATION_KEY = 'relative_to_cluster'

_cache_obj = {
  CACHE_CONFIG_KEY: {},
  CACHE_SVC_KEY: None
}



import os
import subprocess
import traceback
from typing import List, Optional, Dict, Tuple

import yaml
from k8kat.auth.kube_broker import broker
from k8kat.utils.main.api_defs_man import api_defs_man
from typing_extensions import TypedDict

from kama_sdk.core.core import utils
from kama_sdk.core.core.config_man import config_man
from kama_sdk.core.core.types import K8sResDict, KteaDict, KAO

tmp_file_mame = '/tmp/man.yaml'
ktl_apply_cmd_base = f"kubectl apply -f {tmp_file_mame}"
RESD = K8sResDict
RESDs = List[K8sResDict]

class KteaClientConstructorArgs(TypedDict):
  ktea: KteaDict


class KteaClient:

  def __init__(self, **kwargs: KteaClientConstructorArgs):
    self.ktea: KteaDict = kwargs.get('ktea') or config_man.ktea_desc()

  def load_default_values(self):
    raise NotImplemented

  def template_manifest(self, values: Dict) -> List[K8sResDict]:
    raise NotImplemented

  def template_manifest_std(self) -> List[K8sResDict]:
    value_bundle = config_man.standard_manifest_values()
    return self.template_manifest(value_bundle)

  def dry_run(self, **kwargs) -> List[Dict]:
    """
    Retrieves the manifest from Kteai, writes its contents to a temporary local
    file (filtering resource by rules), and runs kubectl apply -f on it.
    :return: any generated terminal output from kubectl apply.
    """
    res_dicts = self.template_manifest(kwargs['values'])
    return self.filter_res(res_dicts, kwargs.get('rules', []))

  def apply(self, **kwargs) -> List[KAO]:
    """
    Retrieves the manifest from Kteai, writes its contents to a temporary local
    file (filtering resource by rules), and runs kubectl apply -f on it.
    :return: any generated terminal output from kubectl apply.
    """
    return self.kubectl_apply(self.dry_run(**kwargs))

  def any_cmd_args(self) -> str:
    return self.ktea.get('args') or ''

  def template_cmd_args(self, vars_path: str) -> str:
    values_flag: str = f"-f {vars_path}"
    ns = config_man.ns()
    return f"{ns} . {values_flag} {self.any_cmd_args()}"

  @staticmethod
  def merge_ktea_vals(spec: KteaDict, values: Dict) -> Dict:
    return utils.deep_merge(
      spec.get('weak_merge', {}),
      values,
      spec.get('strong_merge', {})
    )

  @staticmethod
  def kubectl_dry_run(res_dicts: RESDs) -> Tuple[bool, List[str]]:
    with short_lived_resfile(resdicts=res_dicts):
      try:
        command_parts = ktl_apply_cmd().split(" ")
        result = subprocess.check_output(command_parts, stderr=subprocess.STDOUT)
        return True, bytes2logs(result)
      except subprocess.CalledProcessError as e:
        return False, bytes2logs(e.output)
      except:
        return False, traceback.format_exc()

  @staticmethod
  def kubectl_apply(res_dicts: RESDs) -> List[KAO]:
    """
    Kubectl applies the manifest and returns any generated terminal output.
    :return: any generated teminal output.
    """
    outcomes: List[KAO] = []
    for res_dict in res_dicts:
      with short_lived_resfile(resdict=res_dict):
        command_parts = ktl_apply_cmd().split(" ")
        # noinspection PyBroadException
        try:
          result = subprocess.check_output(command_parts, stderr=subprocess.STDOUT)
          # print(f"[kama_sdk:ktea_client] {command_parts} -----> {result}")
          outcomes.append(log2outcome(True, res_dict, result))
        except subprocess.CalledProcessError as e:
          print(traceback.format_exc())
          print(f"[kama_sdk:ktea_client] CalledProcessError {command_parts} ^^")
          outcomes.append(log2outcome(False, res_dict, e.output))
        except:
          print(traceback.format_exc())
          print(f"[kama_sdk:ktea_client] Unknown error {command_parts} ^^")
          outcomes.append(log2outcome(False, res_dict, traceback.format_exc()))
    return [o for o in outcomes if o]

  @staticmethod
  def filter_res(res_list: RESDs, selectors) -> RESDs:
    """
    Filters the list of parsed kubernetes resource from the tami-generated
    application manifest according to the passed rule-set.
    :param res_list: k8s resource list to be filtered.
    :param selectors: rules to be used for filtering.
    :return: filtered resource list.
    """
    def decide_res(res):
      for selector in selectors:
        if selector.selects_res(res):
          return True
      return False
    if selectors:
      return list(filter(decide_res, res_list))
    else:
      return res_list

  @staticmethod
  def release_name():
    return config_man.ns()


def bytes2logs(output) -> List[str]:
  raw_log = output.decode('utf-8') if output else ''
  parts = raw_log.split("\n")
  # print(f"parts {parts}")
  # raw_log = parts[0] if len(parts) == 2 else None
  # print(f"rawlog {raw_log}")
  return parts


def log2outcome(succs: bool, resdict: RESD, output) -> Optional[KAO]:
  raw_log = output.decode('utf-8') if output else ''
  parts = raw_log.split("\n")
  raw_log = parts[0] if len(parts) == 2 else None

  if raw_log:
    if succs:
      return utils.log2kao(raw_log)
    else:
      kind = resdict.get('kind')
      kind = api_defs_man.kind2plurname(kind) or kind
      api_group = api_defs_man.find_api_group(kind)
      return KAO(
        api_group=api_group,
        kind=kind,
        name=resdict.get('metadata', {}).get('name'),
        verb=None,
        error=raw_log
      )
  else:
    print(f"[kama_sdk::ktea_client] panic log fmt unknown: {raw_log}")
    return None


class short_lived_resfile:
  def __init__(self, **kwargs):
    resdict: K8sResDict = kwargs.get('resdict')
    resdicts: List[K8sResDict] = kwargs.get('resdicts')
    self.res_dict = resdict
    self.res_dicts = resdicts

  def __enter__(self):
    with open(tmp_file_mame, 'w') as file:
      if self.res_dict is not None:
        file.write(yaml.dump(self.res_dict))
      elif self.res_dicts is not None:
        file.write(yaml.dump_all(self.res_dicts))

  def __exit__(self, exc_type, exc_val, exc_tb):
    if os.path.isfile(tmp_file_mame):
      if utils.is_prod():
        os.remove(tmp_file_mame)


def ktl_apply_cmd() -> str:
  final_cmd = ktl_apply_cmd_base
  if utils.is_out_of_cluster():
    if broker.connect_config.get('context'):
      context_part = f"--context={broker.connect_config['context']}"
      final_cmd = f"{final_cmd} {context_part}"
  return final_cmd



from kama_sdk.core.core import utils

"""
Five types of recordable telem events:
  - operation
  - action
  - injection
  - update
  - system check

action.parent = events.*


"""
class TelemEvent:

  def __init(self, **kwargs):
    self.parent_id = kwargs.get('parent_id')
    self.id = utils.rand_str()
    self._type = kwargs.get()


import argparse

from k8kat.auth.kube_broker import broker

from kama_sdk import worker, opsim, server, shell
from kama_sdk.core.core import config_man_helper, utils, config_man
from kama_sdk.model.base import static_validator
from kama_sdk.model.base.model import models_man


def collect_args() -> argparse.Namespace:
  parser = argparse.ArgumentParser(description='Process some integers.')
  parser.add_argument(
    "-m",
    "--mode",
    default='server',
    choices=["server", "worker", "telem_worker", "shell", "opsim", 'validation'],
    help="which type of process should start"
  )

  parser.add_argument(
    "-ns",
    "--namespace",
    help="coerce a kubernetes namespace during development"
  )

  parser.add_argument(
    "--validate",
    default='true',
    choices=['true', 'false', 'yes', 'no']
  )

  return parser.parse_args()


def process_coerce_ns_request(args: argparse.Namespace):
  if ns := args.namespace:
    if not utils.is_in_cluster():
      config_man.coerce_ns(ns)
      print(f"[kama_sdk:entrypoint] coerced namespace to {ns}")
    else:
      print(f"[kama_sdk:entrypoint] cannot coerce namespace while in-cluster")


def start():
  args = collect_args()
  broker.connect()
  models_man.add_defaults()
  config_man_helper.clear_trackers()
  process_coerce_ns_request(args)

  if args.mode != 'validation' and utils.any2bool(args.validate):
    static_validator.run_all()

  if args.mode == 'server':
    server.start()
  elif args.mode == 'worker':
    worker.start_main()
  elif args.mode == 'telem_worker':
    worker.start_telem()
  elif args.mode == 'opsim':
    opsim.start()
  elif args.mode == 'shell':
    shell.start()
  elif args.mode == 'validation':
    static_validator.run_all()
  else:
    print(f"Unrecognized exec mode {args.mode}")



from typing import List, Dict

from werkzeug.utils import cached_property

from kama_sdk.core.core import utils
from kama_sdk.model.supplier.base.supplier import Supplier, SRC_DATA_KEY


class Redactor(Supplier):

  @cached_property
  def blacklist(self) -> List[str]:
    return self.get_prop(BLACKLIST_KEY, [])

  @cached_property
  def whitelist(self):
    return self.get_prop(WHITELIST_KEY, [])

  def source_data(self) -> Dict:
    return self.resolve_prop(SRC_DATA_KEY, depth=10, backup={})

  def _compute(self) -> Dict:
    original = self.source_data()
    if type(original) == dict:
      flat_original = utils.deep2flat(original)
      flat_keys = flat_original.keys()
      filtered_flat_keys = flat_keys

      if self.whitelist:
        filtered_flat_keys = [k for k in flat_keys if k in self.whitelist]

      if self.blacklist:
        filtered_flat_keys = [k for k in flat_keys if k not in self.blacklist]

      filtered_flat_dict = {k: flat_original[k] for k in filtered_flat_keys}
      return utils.flat2deep(filtered_flat_dict)
    else:
      return original


BLACKLIST_KEY = 'blacklist'
WHITELIST_KEY = 'whitelist'



from typing import Optional

from k8kat.res.svc.kat_svc import KatSvc
from kubernetes.client import V1EndpointAddress, V1ObjectReference
from werkzeug.utils import cached_property

from kama_sdk.core.core import utils
from kama_sdk.core.core.types import EndpointDict, PortForwardSpec
from kama_sdk.model.supplier.base.supplier import Supplier
from kama_sdk.model.supplier.ext.biz.resource_selector import ResourceSelector


class ServiceEndpointSupplier(Supplier):

  def _compute(self) -> Optional[EndpointDict]:
    if svc := self.underlying_svc:
      port_part = f":{self.port}" if self.port else ''
      available_ip = svc.external_ip or svc.internal_ip
      forwarding_spec = self.port_fwd_spec() if not svc.external_ip else None
      return EndpointDict(
        url=f"{available_ip}{port_part}",
        svc_type=svc.type,
        port_forward_spec=forwarding_spec
      )
    else:
      return None

  @cached_property
  def port_fwd_if_internal(self) -> str:
    return self.get_prop(PORT_FWD_IF_INTERNAL_KEY, True)

  @cached_property
  def resource_selector(self):
    return self.inflate_child(
      ResourceSelector,
      prop=RESOURCE_KEY,
      safely=True
    )

  @cached_property
  def underlying_svc(self) -> Optional[KatSvc]:
    pre_defined = self.source_data()
    if pre_defined and type(pre_defined) == KatSvc:
      return pre_defined
    else:
      if self.resource_selector:
        matches = self.resource_selector.query_cluster()
        res = next(iter(matches), None)
        if res and isinstance(res, KatSvc):
          return res
      return None

  @cached_property
  def port(self) -> Optional[str]:
    return self.get_prop(PORT_KEY) or self.infer_port()

  def infer_url(self) -> Optional[str]:
    return self.underlying_svc.external_ip or \
           self.underlying_svc.internal_ip

  def infer_port(self) -> str:
    value = str(self.underlying_svc.first_tcp_port_num())
    return value if not value == '80' else ''

  def port_fwd_spec(self) -> Optional[PortForwardSpec]:
    winner = None
    backend_dicts = self.underlying_svc.flat_endpoints() or []
    for backend_dict in utils.compact(backend_dicts):
      address: V1EndpointAddress = backend_dict
      target_ref: V1ObjectReference = address.target_ref
      if target_ref and target_ref.kind == 'Pod':
        winner = target_ref
        break

    if winner:
      return PortForwardSpec(
        namespace=winner.namespace,
        pod_name=winner.name,
        pod_port=int(self.port or '80')
      )


RESOURCE_KEY = 'selector'
PORT_KEY = 'port'
PORT_FWD_IF_INTERNAL_KEY = 'port_forward_if_internal'



import os

from flask import Flask, jsonify, request
from flask_cors import CORS
from k8kat.auth.kube_broker import BrokerConnException

from kama_sdk.controllers import operations_controller, status_controller, \
  app_controller, manifest_variables_controller, updates_controller, \
  telem_controller, injections_controller, system_checks_controller, concerns_controller
from kama_sdk.core.core import utils
from kama_sdk.core.core.config_man import coerce_ns

app = Flask(__name__)
app.config["SECRET_KEY"] = os.environ.get('SECRET_KEY')

controllers = [
  status_controller,
  operations_controller,
  app_controller,
  concerns_controller,
  telem_controller,
  updates_controller,
  manifest_variables_controller,
  injections_controller,
  system_checks_controller
]

for controller in controllers:
  app.register_blueprint(controller.controller)


CORS(app)


@app.errorhandler(BrokerConnException)
def all_exception_handler(error):
  return jsonify(dict(
    error='could not connect to Kubernetes API',
    reason=str(error)
  )), 500


@app.before_request
def read_dev_ns():
  coerced_ns = request.headers.get('Appns')
  if coerced_ns:
    if utils.is_out_of_cluster():
      coerce_ns(coerced_ns)
    else:
      print(f"[kama_sdk::server] set-ns from header env={utils.run_env()}!")


def start():
  app.config["cmd"] = ["bash"]
  app.config['JSON_SORT_KEYS'] = False
  app.run(host='0.0.0.0', port=5000, extra_files=['./**/*.yaml'])



import os
from copy import deepcopy
from os.path import isfile
from typing import Type, Optional, Dict, Union, List, TypeVar, Any, Tuple

import yaml
# noinspection PyUnresolvedReferences
from werkzeug.utils import cached_property, invalidate_cached_property

from kama_sdk.core.core import utils, consts, str_sub
from kama_sdk.core.core.types import KoD
from kama_sdk.core.core.utils import perr, pwar
from kama_sdk.model.base.default_models import default_model_classes
from kama_sdk.model.base.exceptions import InflateModelError

T = TypeVar('T', bound='Model')


ConfigLookup = Tuple[bool, Any]


class ModelsMan:
  def __init__(self):
    self._descriptors: List[Dict] = []
    self._classes: List[Type[T]] = []
    self._asset_paths: List[str] = []

  def add_descriptors(self, descriptors: List[Dict]):
    self._descriptors = self._descriptors + descriptors

  def add_classes(self, model_classes: List[Type[T]]):
    self._classes = self._classes + model_classes

  def add_asset_dir_paths(self, paths: List[str]):
    self._asset_paths += paths

  def add_defaults(self):
    self.add_descriptors(default_descriptors())
    self.add_classes(default_model_classes())
    self.add_asset_dir_paths(default_asset_paths())

  def clear(self, restore_defaults=True):
    self._descriptors = []
    self._classes = []
    if restore_defaults:
      self.add_defaults()

  def descriptors(self) -> List[Dict]:
    return self._descriptors

  def classes(self) -> List[Type[T]]:
    return self._classes

  def asset_dir_paths(self) -> List[str]:
    return self._asset_paths


models_man = ModelsMan()


class Model:

  def __init__(self, config: Dict):
    self.config: Dict = config
    self._id: str = config.get(ID_KEY)
    self.parent: Optional[T] = None
    self.cached_results = {}
    self.config['kind'] = self.__class__.__name__

  def id(self) -> str:
    return self._id

  def sig(self):
    return f"{self.kind()}/{self.id() or self.title}"

  def patch(self, new_props: Dict[str, any]) -> T:
    self.config = {**self.config, **new_props}
    for key in self.config.keys():
      if key in self.__dict__:
        invalidate_cached_property(self, key)
    return self

  def nth_parent(self, n: int) -> T:
    if n == 0:
      return self
    else:
      return self.parent.nth_parent(n - 1) if self.parent else None

  @cached_property
  def title(self) -> Optional[str]:
    return self.resolve_prop(TITLE_KEY, lookback=0)

  @cached_property
  def info(self) -> Optional[str]:
    return self.resolve_prop(INFO_KEY, lookback=0)

  @cached_property
  def tags(self) -> List[str]:
    return self.resolve_prop(TAGS_KEY, lookback=0, backup=[])

  @cached_property
  def default_lookback(self) -> Union[int, bool]:
    return self.config.get(LOOKBACK_KEY, True)

  def to_dict(self):
    return dict(
      id=self.id(),
      title=self.title,
      info=self.info
    )

  def update_attrs(self, config: Dict):
    for key, value in config.items():
      setattr(self, key, value)

  def get_prop(self, key: str, backup: Any = None) -> Any:
    """
    Convenience method for resolve_props
    @param key: name of desired value in config dict
    @param backup: value to return if value not in config dict
    @return: fully resolved property value or backup if given
    """
    return self.resolve_prop(key, backup=backup)

  def resolve_prop(self, key: str, **kwargs) -> Any:
    handled, value = self.do_local_config_lookup(key, **kwargs)

    if handled:
      depth = kwargs.pop('depth', None)
      return self.resolve_prop_value(value, depth) if value else value
    else:
      backup_handled, backup_value = resolve_backup(**kwargs)
      if backup_handled:
        return backup_value
      else:
        if on_miss := kwargs.pop('missing', None):
          msg = f"required prop {key} not supplied"
          error = f"[kama_sdk:{self.__class__.__name__}] {msg}"
          if on_miss == 'warn':
            print(error)
          if on_miss == 'raise':
            raise RuntimeError(error)
        return None

  def do_local_config_lookup(self, key: str, **kwargs) -> ConfigLookup:
    """
    Reads a value from the main config dicts, applies all possible
    resolution transformations to obtain its final value. This includes
    1) IFTT resolution, 2) ValueSupplier resolution, and
    3) string substitutions.
    @param key:
    @param kwargs:
    @return: fully resolved property value or backup if given
    """
    value = None

    prevent_aliasing = kwargs.pop(PREVENT_ALIASING_KEY, False)
    standard_keys = self.config.keys()
    as_new = lambda k: f"{NEW_PROP_PREF}{k}"
    # handled = False

    if prevent_aliasing:
      if handled := key in self.config.keys():
        value = self.config.get(key)
    elif as_new(key) in standard_keys and not key.startswith(NEW_PROP_PREF):
      value = self.config[as_new(key)]
      handled = True
    elif key in standard_keys:
      value = self.config[key]
      handled = True
    elif key.startswith(OLD_PROP_PREF):
      value = self.resolve_prop(
        key.replace(OLD_PROP_PREF, ""),
        **kwargs,
        **{PREVENT_ALIASING_KEY: True},
      )
      handled = True
    else:
      handled, value = self.resolve_from_cache(key)

    if not handled:
      handled, value = self.do_ancestral_config_lookup(
        key,
        **{PREVENT_ALIASING_KEY: prevent_aliasing},
        **kwargs,
      )

    return handled, value

  def user_props_cache(self) -> Dict:
    return self.config.get(CACHED_KEY, {})

  def resolve_from_cache(self, key: str) -> ConfigLookup:
    if key in self.user_props_cache().keys():
      if key in self.cached_results.keys():
        return True, self.cached_results[key]
      else:
        unresolved_value = self.user_props_cache()[key]
        resolved_value = self.resolve_prop_value(unresolved_value)
        # resolved_value = unresolved_value
        self.cached_results[key] = resolved_value
        return True, resolved_value
    else:
      return False, None

  def do_ancestral_config_lookup(self, key: str, **kwargs) -> ConfigLookup:
    explicit_lookback_req = kwargs.pop(LOOKBACK_KEY, None)
    lookback = self.lookback2int(explicit_lookback_req)

    if self.parent and lookback > 0:
      return self.parent.do_local_config_lookup(
        key,
        **{LOOKBACK_KEY: lookback - 1},
        **kwargs
      )
    else:
      return False, None

  def resolve_prop_value(self, value: Any, depth=0) -> Optional[Any]:
    if type(value) == str:
      return self.resolve_str_like_prop_value(value)
    elif type(value) == list:
      return self.resolve_list_like_prop_value(value)
    elif type(value) == dict:
      return self.resolve_dict_like_prop_value(value, depth)
    else:
      return value

  def resolve_str_like_prop_value(self, value: Any):
    if value in NULLISH_ALIASES:
      return None
    else:
      value = self.supplier_resolve_or_identity(value)
      value = self.interpolate_prop(value)
      return self.try_read_from_asset(value)

  def resolve_list_like_prop_value(self, value: List):
    resolved_sub_values = []
    for sub_value in value:
      if type(sub_value) == str and sub_value.startswith(SPLATTER_STR):
        resolvable_part = sub_value.split(SPLATTER_STR)[1]
        resolved_sub_value = self.resolve_prop_value(resolvable_part)
        if utils.listlike(resolved_sub_value):
          resolved_sub_values.extend(resolved_sub_value)
        elif resolved_sub_value is not None:
          tp = type(resolved_sub_value)
          pwar(self, f"danger cannot splatter {tp} {resolved_sub_value}")
      else:
        resolved_sub_value = self.resolve_prop_value(sub_value)
        resolved_sub_values.append(resolved_sub_value)
    return resolved_sub_values

  def resolve_dict_like_prop_value(self, value: Dict, depth=0):
    txd_value = self.supplier_resolve_or_identity(value)
    if type(txd_value) == dict and int(depth or 0) > 0:
      perf = lambda v: self.resolve_prop_value(v, depth - 1)
      return {k: perf(v) for (k, v) in txd_value.items()}
    else:
      return txd_value

  def interpolate_prop(self, value: str) -> Any:
    """
    Performs string substitutions on input. Combines substitution context
    from instance's self.context and any additional context passed as
    parameters. Returns unchanged input if property is not a string.
    @param value: value of property to interpolate
    @return: interpolated string if input is string else unchanged input
    """
    if value and type(value) == str:
      def resolve(expr: str):
        ret = self.supplier_resolve_or_identity(expr)
        return str(ret) if ret else ''
      return str_sub.interpolate(value, resolve)
    else:
      return value

  @classmethod
  def kind(cls):
    return cls.__name__

  def serialize(self) -> Dict[str, Any]:
    src_items = list(self.config.items())
    no_copy = ['cache']
    return {k: v for k, v in src_items if k not in no_copy}

  def clone(self) -> T:
    new_model = self.__class__(self.serialize())
    new_model.parent = self.parent
    return new_model

  def inflate_children(self, child_class: Type[T], **kwargs):
    """
    Bottleneck function for a parent model to inflate a list of children.
    In the normal case, kods_or_provider_kod is a list of WizModels KoDs.
    In the special case, kods_or_provider_kod is ListGetter model
    that produces the actual children.
    case,
    @param child_class: class all children must a subclass of
    @return: resolved list of WizModel children
    """
    kods_or_supplier = self.resolve_child_ref_kod(
      kod=kwargs.pop('kod', None),
      prop=kwargs.pop('prop', None),
      resolve_kod=kwargs.pop(RESOLVE_CHILD_KOD_FLAG, True)
    ) or []

    to_child = lambda obj: self.kod2child(obj, child_class, **kwargs)

    if utils.listlike(kods_or_supplier):
      children_kods: List[KoD] = kods_or_supplier
      return list(map(to_child, children_kods))
    elif type(kods_or_supplier) in [str, dict]:
      from kama_sdk.model.supplier.base.supplier import Supplier
      # children_kods = self.resolve_prop_value(kods_or_supplier)
      children_kods = self.supplier_resolve_or_identity(kods_or_supplier)
      if type(children_kods) == list:
        return list(map(to_child, children_kods))
      else:
        err = f"children must be list or supplier not {kods_or_supplier}"
        raise RuntimeError(f"[Model] {err}")

  def inflate_child(self, child_cls: Type[T], **kwargs) -> Optional[T]:
    child_kod: KoD = self.resolve_child_ref_kod(
      kod=kwargs.pop('kod', None),
      prop=kwargs.pop('prop', None),
      resolve_kod=kwargs.pop(RESOLVE_CHILD_KOD_FLAG, True)
    )
    return self.kod2child(child_kod, child_cls, **kwargs)

  def resolve_child_ref_kod(self, **kwargs) -> Union[List[KoD], KoD]:
    if explicit_kod := kwargs.get('kod'):
      return explicit_kod
    elif prop_name := kwargs.get('prop'):
      if kwargs.get(RESOLVE_CHILD_KOD_FLAG, True):
        return self.get_prop(prop_name)
      else:
        return self.config.get(prop_name)
    else:
      base = f"child ref from {self.sig()} must include kod or prop"
      perr(self, f"{base} , neither found in {kwargs}")

  def kod2child(self, kod: KoD, child_cls: Type[T], **kwargs) -> T:
    try:
      down_kw = {**kwargs, 'mute': True}
      inflated = child_cls.inflate(kod, **down_kw)
      if inflated:
        inflated.parent = self
      return inflated
    except Exception as e:
      exp_child_kind = child_cls.__name__ if isinstance(child_cls, type) else '?'
      child_part = f"child {exp_child_kind} = {kod}"
      perr(self, f"parent {self.sig()} failed to inflate {child_part}")
      raise e

  def prop_inheritance_pool(self):
    return self.config

  def lookback2int(self, override: Union[bool, int]) -> int:
    overridden = override is not None
    lookback = override if overridden else self.default_lookback
    if type(lookback) == int:
      return lookback
    else:
      return 1000 if lookback else 0

  @classmethod
  def inflate_all(cls, **kwargs) -> List[T]:
    cls_pool = cls.lteq_classes(models_man.classes())
    configs = configs_for_kinds(models_man.descriptors(), cls_pool)
    return [cls.inflate(config, **kwargs) for config in configs]

  @classmethod
  def inflate_many(cls: T, kods: List[KoD], **kwargs) -> List[T]:
    instances: List[T] = []
    safely = kwargs.pop('safely', False)
    for kod in kods:
      try:
        instances.append(cls.inflate(kod, **kwargs))
      except Exception as e:
        if not safely:
          raise e
    return instances

  @classmethod
  def inflate(cls: T, kod: KoD, **kwargs) -> Optional[T]:
    try:
      if isinstance(kod, str):
        return cls.inflate_with_str(kod, **kwargs)
      elif isinstance(kod, Dict):
        return cls.inflate_with_config(kod, **kwargs)
      else:
        raise InflateModelError(f"inflate value {kod} neither dict or str")
    except Exception as err:
      if kwargs.pop('safely', False):
        return None
      else:
        if not kwargs.pop('mute', False):
          perr(cls, f"inflate error below for {kod}")
        raise err

  @classmethod
  def from_yaml(cls, yaml_str: str, **kwargs) -> Optional[T]:
    as_dict = yaml.load(yaml_str)
    return cls.inflate_with_config(as_dict, **kwargs)

  @classmethod
  def inflate_with_str(cls, string: str, **kwargs) -> T:
    if string.startswith(KIND_REFERENCE_PREFIX):
      referenced_kind = string.split(KIND_REFERENCE_PREFIX)[1]
      return cls.inflate_with_kind_reference(referenced_kind, **kwargs)

    elif string.startswith(ID_REFERENCE_PREFIX):
      referenced_id = string.split(ID_REFERENCE_PREFIX)[1]
      return cls.inflate_with_id_reference(referenced_id, **kwargs)

    elif string.startswith(EXPR_REFERENCE_PREFIX):
      literal_expr = string.split(EXPR_REFERENCE_PREFIX)[1]
      return cls.inflate_with_literal(literal_expr, **kwargs)

    else:
      return cls.inflate_with_id_reference(string, **kwargs)

  @classmethod
  def inflate_with_literal(cls, expr: str, **kwargs) -> Optional[T]:
    return None

  @classmethod
  def inflate_with_kind_reference(cls, kind_ref: str, **kwargs) -> T:
    config = dict(kind=kind_ref)
    return cls.inflate_with_config(config, **kwargs)

  @classmethod
  def inflate_with_id_reference(cls, id_ref: str, **kwargs) -> T:
    candidate_subclasses = cls.lteq_classes(models_man.classes())
    candidate_kinds = [klass.kind() for klass in candidate_subclasses]
    all_configs = models_man.descriptors()
    config = find_config_by_id(id_ref, all_configs, candidate_kinds)
    return cls.inflate_with_config(config, **kwargs)

  @classmethod
  def descendent_or_self(cls) -> T:
    subclasses = cls.lteq_classes(models_man.classes())
    not_self = lambda kls: not kls == cls
    return next(filter(not_self, subclasses), cls)({})

  @classmethod
  def inflate_with_config(cls, config: Dict, **kwargs) -> T:

    if config is None:
      raise InflateModelError("inflate_with_config given no config")

    patches: Optional[Dict] = kwargs.get('patches', {})

    host_class = cls

    inherit_id = config.get(INHERIT_KEY)
    explicit_kind = config.get(KIND_KEY)

    if explicit_kind and not explicit_kind == host_class.__name__:
      host_class = cls.kind2cls(explicit_kind)
      if not host_class:
        err = f"no kind {explicit_kind} under {cls.__name__}"
        raise RuntimeError(f"[Model] FATAL {err}")

    if inherit_id:
      other: T = cls.inflate_with_str(inherit_id, **kwargs)
      host_class = other.__class__
      # config = {**other.serialize(), **config}
      config = utils.deep_merge(other.serialize(), config)

    final_config = deepcopy(utils.deep_merge(config, patches))
    model_instance = host_class(final_config)

    return model_instance

  @classmethod
  def global_provider_id(cls):
    raise NotImplemented

  @classmethod
  def from_provider(cls) -> List[T]:
    from kama_sdk.model.supplier.base.supplier import Supplier
    provider: Supplier = Supplier.inflate(cls.global_provider_id())
    return cls.inflate_many(provider.resolve())

  @classmethod
  def singleton_id(cls):
    raise NotImplemented

  @classmethod
  def inflate_singleton(cls, **kwargs) -> T:
    return cls.inflate_with_str(cls.singleton_id(), **kwargs)

  def supplier_resolve_or_identity(self, kod: KoD, **kwargs) -> Any:
    from kama_sdk.model.supplier.base.supplier import Supplier
    prefix: str = "get::"

    if self.is_interceptor_candidate(Supplier, prefix, kod):
      final_kod = kod
      if type(kod) == str:
        trimmed_kod = kod.replace(prefix, "")
        final_kod = Supplier.expr2props(trimmed_kod)

      kwargs.pop('safely', None)
      interceptor = Supplier.inflate(
        final_kod,
        safely=True,
        **kwargs
      )
      if interceptor:
        interceptor.parent = self
        return interceptor.resolve()

    return kod

  @classmethod
  def id_exists(cls, _id: str) -> bool:
    return True

  @staticmethod
  def truncate_kod_prefix(kod: KoD, prefix: str) -> KoD:
    if type(kod) == str and len(kod) >= len(prefix):
      return kod[len(prefix):len(kod)]
    return kod

  @classmethod
  def is_interceptor_candidate(cls, interceptor: Type[T], prefix, kod: KoD):
    if type(kod) == dict:
      interceptors = interceptor.lteq_classes(models_man.classes())
      if kod.get('kind') in [c.__name__ for c in interceptors]:
        return True
    if type(kod) == str:
      return kod.startswith(prefix)
    return False

  @classmethod
  def lteq_classes(cls, classes: List[Type]) -> List[Type[T]]:
    return [klass for klass in [*classes, cls] if issubclass(klass, cls)]

  @classmethod
  def kind2cls(cls, kind: str):
    subclasses = cls.lteq_classes(models_man.classes())
    return find_class_by_name(kind, subclasses)

  @staticmethod
  def try_read_from_asset(value):
    if value and type(value) == str:
      if value.startswith(ASSET_PREFIX):
        return read_from_asset(value)
    return value


def read_from_asset(descriptor: str) -> str:
  _, path = descriptor.split("::")
  for dirpath in models_man.asset_dir_paths():
    full_path = f"{dirpath}/{path}"
    if isfile(full_path):
      with open(full_path) as file:
        return file.read()
  return ''


def key_or_dict_to_key(key_or_dict: Union[str, dict]) -> str:
  if isinstance(key_or_dict, str):
    return key_or_dict
  elif isinstance(key_or_dict, dict):
    return key_or_dict.get('id')
  raise RuntimeError(f"Can't handle {key_or_dict}")


def key_or_dict_matches(key_or_dict: KoD, target_key: str) -> bool:
  return key_or_dict_to_key(key_or_dict) == target_key


def find_class_by_name(name: str, classes) -> Type:
  matcher = lambda klass: klass.__name__ == name
  return next(filter(matcher, classes), None)

def find_any_class_by_name(name: str) -> Type:
  matcher = lambda klass: klass.__name__ == name
  return next(filter(matcher, models_man.classes()), None)

def find_any_config_by_id(_id: str) -> Dict:
  matcher = lambda c: c.get('id') == _id
  return next(filter(matcher, models_man.descriptors()), None)

def find_config_by_id(_id: str, configs: List[Dict], kinds: List[str]) -> Dict:
  matcher = lambda c: c.get('id') == _id and c.get('kind') in kinds
  return next(filter(matcher, configs), None)


def configs_for_kinds(configs: List[Dict], cls_pool) -> List[Dict]:
  kinds_pool = [cls.__name__ for cls in cls_pool]
  return [c for c in configs if c.get('kind') in kinds_pool]


def default_descriptors() -> List[Dict]:
  pwd = os.path.join(os.path.dirname(__file__))
  search_space_root = f"{pwd}/../../model"
  yaml_dirs = discover_model_yaml_dirs(search_space_root)
  return utils.flatten(list(map(utils.yamls_in_dir, yaml_dirs)))


def discover_model_yaml_dirs(root: str) -> List[str]:
  pwd = root.split('/')[-1]
  is_root_yaml_dir = pwd == 'yamls'
  child_paths = [f"{root}/{name}" for name in os.listdir(root)]
  child_dirs = list(filter(os.path.isdir, child_paths))
  child_results = list(map(discover_model_yaml_dirs, child_dirs))
  self_result = [root] if is_root_yaml_dir else []
  return self_result + utils.flatten(child_results)


def default_asset_paths() -> List[str]:
  pwd = os.path.join(os.path.dirname(__file__))
  return [f"{pwd}/../../assets"]

def resolve_backup(**kwargs):
  lazy_backup = kwargs.pop('lazy_backup', None)
  if lazy_backup is not None:
    return True, lazy_backup()
  elif 'backup' in kwargs.keys():
    return True, kwargs.pop('backup', None)
  else:
    return False, None


NULLISH_ALIASES = ['__none__', '__null__', '__nil__', '__undefined__']
NEW_PROP_PREF = "new__"
OLD_PROP_PREF = "old__"
ID_REFERENCE_PREFIX = "id::"
EXPR_REFERENCE_PREFIX = "expr::"
KIND_REFERENCE_PREFIX = "kind::"
PREVENT_ALIASING_KEY = 'old'
LOOKBACK_KEY = 'lookback'
SPLATTER_STR = '...'
RESOLVE_CHILD_KOD_FLAG = 'resolve_kod'
CACHED_KEY = 'cached'

ID_KEY = 'id'
TITLE_KEY = 'title'
INHERIT_KEY = 'inherit'
KIND_KEY = 'kind'
INFO_KEY = 'info'
TAGS_KEY = 'tags'
ASSET_PREFIX = f'{consts.FILE_PREFIX}::'



from typing import List

from kama_sdk.core.core import utils
from kama_sdk.core.core.types import KDLoS
from kama_sdk.model.supplier.base import supplier
from kama_sdk.model.supplier.predicate.predicate import Predicate, OPERATOR_KEY


class MultiPredicate(Predicate):

  def sub_predicates(self) -> List[KDLoS]:
    return self.resolve_prop(supplier.SRC_DATA_KEY, missing='warn')

  def operator(self):
    return self.get_prop(OPERATOR_KEY, 'and')

  def resolve(self) -> bool:
    operator = self.operator()
    resolvable_predicates = self.sub_predicates()
    for sub_pred in resolvable_predicates:
      eval_or_literal = self.resolve_prop_value(sub_pred)
      resolved_to_true = utils.any2bool(eval_or_literal)

      if operator == 'or':
        if resolved_to_true:
          return True
      elif operator == 'and':
        if not resolved_to_true:
          return False
      else:
        print(f"[kama_sdk::multi_pred] illegal operator {operator}")
        return False
    return operator == 'and'



from kama_sdk.core.core import consts
from kama_sdk.core.core.config_man import config_man
from kama_sdk.core.ktea.http_ktea_client import HttpKteaClient
from kama_sdk.core.ktea.ktea_client import KteaClient
from kama_sdk.core.ktea.local_exec_ktea_client import LocalExecKteaClient
from kama_sdk.core.ktea.virtual_ktea_client import vktea_clients_man


def ktea_client(**kwargs) -> KteaClient:
  ktea = kwargs.pop('ktea', None) or config_man.ktea_desc()
  ktea_type = ktea['type']
  # print(f"MOVING FWD WITH KTEA {ktea}")

  if ktea_type == consts.http_api_ktea:
    return HttpKteaClient(ktea=ktea, **kwargs)
  elif ktea_type == consts.local_exec_ktea:
    return LocalExecKteaClient(ktea=ktea, **kwargs)
  elif ktea_type == consts.virtual_ktea:
    klass_or_name = ktea['uri']
    if type(klass_or_name) == type:
      klass = klass_or_name
    elif type(klass_or_name) == str:
      klass = vktea_clients_man.find_client(name=klass_or_name)
    else:
      raise Exception(f"[ktea_provider] invalid ktea uri {klass_or_name}")
    return klass(**kwargs)
  else:
    raise RuntimeError(f"Illegal KTEA type {ktea_type}")



from datetime import datetime
from typing import Dict, Optional

from kama_sdk.core.core import consts
from kama_sdk.core.core.types import ExitStatuses, ActionOutcome

IDLE = 'idle'
RUNNING = 'running'
SETTLED_POS = 'positive'
SETTLED_NEG = 'negative'


class StepState:
  def __init__(self, step_sig: str, parent_op):
    self.step_sig: str = step_sig
    self.parent_op = parent_op
    self.started_at = datetime.now()
    self.chart_assigns: Dict = {}
    self.state_assigns: Dict = {}
    self.pref_assigns: Dict = {}
    self.action_outcome: Optional[ActionOutcome] = None
    self.action_telem = None
    self.exit_statuses: ExitStatuses = default_exit_statuses()
    self.committed_at = None
    self.terminated_at = None
    self.job_id = None

  def notify_vars_assigned(self, bundle: Dict):
    self.chart_assigns = bundle.get(consts.TARGET_STANDARD)
    self.state_assigns = bundle.get(consts.TARGET_STATE)
    self.pref_assigns = bundle.get(consts.TARGET_PREFS)

  def all_assigns(self):
    """
    Merges chart assigns and state assigns extracted from the commit outcome.
    :return:
    """
    return dict(
      **self.chart_assigns,
      **self.state_assigns,
      **self.pref_assigns
    )

  def persistable_assigns(self) -> Dict:
    return dict(
      **self.chart_assigns,
      **self.pref_assigns
    )

def default_exit_statuses() -> ExitStatuses:
  return ExitStatuses(positive=[], negative=[])



import code
import os
from typing import Optional, Dict

import yaml
from k8kat.res.base.kat_res import KatRes
from k8kat.res.config_map.kat_map import KatMap
from k8kat.res.dep.kat_dep import KatDep
from k8kat.res.node.kat_node import KatNode
from k8kat.res.ns.kat_ns import KatNs
from k8kat.res.pod.kat_pod import KatPod
from k8kat.res.svc.kat_svc import KatSvc

from kama_sdk.core.core import utils, updates_man
from kama_sdk.core.core.config_man import config_man
from kama_sdk.core.core import prom_api_client as prom_client
from kama_sdk.core.telem import telem_man
from kama_sdk.model.base.model import models_man, Model


def models_context():
  return {c.__name__: c for c in models_man.classes()}


def k8kat_context():
  return {c.__name__: c for c in kat_res_classes}


def misc_context():
  return {
    'Model': Model,
    'utils': utils,
    'r': load_dict_from_scratch,
    'config_man': config_man,
    'models_man': models_man,
    'updates_man': updates_man,
    'telem_man': telem_man,
    'use_prom': use_prom
  }


def start():
  context = {
    **models_context(),
    **k8kat_context(),
    **misc_context()
  }

  setup_console_history()
  code.interact(None, None, context, None)


def setup_console_history():
  import os as loc_os
  import atexit
  import readline
  import rlcompleter

  history_path = os.path.expanduser("~/.pyhistory")

  def save_history(_history_path=history_path):
    import readline
    readline.write_history_file(_history_path)

  if os.path.exists(history_path):
    readline.read_history_file(history_path)

  atexit.register(save_history)
  del loc_os, atexit, readline, rlcompleter, save_history, history_path


def load_dict_from_scratch(_id: str = None) -> Optional[Dict]:
  scratch_file_path = f"{os.getcwd()}/scratch.yaml"
  with open(scratch_file_path, 'r') as scratch_file:
    model_dicts = list(yaml.full_load_all(scratch_file.read()))
    if _id:
      finder = lambda d: d.get('id') == _id
      if res := next(filter(finder, model_dicts), None):
        return res
      else:
        print(f"[kama_sdk:shell] {_id} not found in {scratch_file_path}")
    else:
      if len(model_dicts) > 0:
        return model_dicts[0]
      else:
        print(f"[kama_sdk:shell] scratch {scratch_file_path} has 0 defs")


def use_prom():
  config_man.patch_prefs({
    'monitoring': {
      prom_client.LOCATION_KEY: 'in',
      prom_client.IS_PROXY_KEY: True,
      prom_client.SVC_NS_KEY: 'monitoring',
      prom_client.SVC_NAME_KEY: 'monitoring-kube-prometheus-prometheus'
    }

  })


kat_res_classes = [
  KatRes,
  KatPod,
  KatDep,
  KatSvc,
  KatMap,
  KatNs,
  KatNode
]



from typing import Any, Dict, Union, Optional

import jq
from werkzeug.utils import cached_property

from kama_sdk.core.core import utils
from kama_sdk.core.core.utils import listlike
from kama_sdk.model.base.model import Model


class Supplier(Model):

  @cached_property
  def treat_as_list(self) -> bool:
    value = self.resolve_prop(IS_MANY_KEY, lookback=0)
    return value

  @cached_property
  def output_format(self):
    return self.resolve_prop(OUTPUT_FMT_KEY, lookback=0)

  def source_data(self) -> Optional[any]:
    return self.resolve_prop(SRC_DATA_KEY, lookback=0)

  @cached_property
  def serializer_type(self) -> str:
    return self.resolve_prop(
      SERIALIZER_KEY,
      backup='jq',
      lookback=0
    )

  def resolve(self) -> Any:
    computed_value = self._compute()
    if self.serializer_type == SER_LEG:
      return self.serialize_computed_value_legacy(computed_value)
    elif self.serializer_type == SER_JQ:
      return self.jq_serialize(computed_value)
    elif self.serializer_type == SER_IDENTITY:
      return computed_value
    else:
      pt1 = f"treating unknown ser {self.serializer_type}"
      print(f"[kama_sdk:supplier] {pt1} as {SER_IDENTITY}")
      return computed_value

  def jq_serialize(self, value: Any) -> any:
    if self.output_format and value is not None:
      try:
        expression = jq.compile(self.output_format)
        if self.treat_as_list:
          return expression.input(value).all()
        else:
          return expression.input(value).first()
      except Exception as e:
        print(f"[kama_sdk] danger JQ compile failed: {str(e)}")
        return None
    else:
      return value

  def serialize_computed_value_legacy(self, computed_value) -> Any:
    treat_as_list = self.treat_as_list
    if treat_as_list in [None, 'auto']:
      treat_as_list = listlike(computed_value)

    if treat_as_list and not self.output_format == '__count__':
      if listlike(computed_value):
        return [self.serialize_item(item) for item in computed_value]
      else:
        return [self.serialize_item(computed_value)]
    else:
      if not listlike(computed_value) or self.output_format == '__count__':
        return self.serialize_item(computed_value)
      else:
        item = computed_value[0] if len(computed_value) > 0 else None
        return self.serialize_item(item) if item else None

  def _compute(self) -> Any:
    return self.source_data()

  def serialize_item(self, item: Any) -> Union[Dict, str]:
    fmt = self.output_format
    if not fmt or type(fmt) == str:
      return self.serialize_item_prop(item, fmt)
    elif type(fmt) == dict:
      return self.serialize_dict_item(item)
    else:
      return ''

  def serialize_dict_item(self, item):
    fmt: Dict = self.output_format
    serialized = {}
    for key, value in list(fmt.items()):
      serialized[key] = self.serialize_item_prop(item, value)
    return serialized

  # noinspection PyBroadException
  @staticmethod
  def serialize_item_prop(item: Any, prop_name: Optional[str]) -> Optional[Any]:
    if prop_name:
      if prop_name == '__identity__':
        return item
      elif prop_name == '__count__':
        try:
          return len(item)
        except:
          return 0
      else:
        try:
          return utils.pluck_or_getattr_deep(item, prop_name)
        except:
          return None
    else:
      return item

  @classmethod
  def expr2props(cls, expr: str) -> Dict:
    parts = expr.split(" ")
    identity_expr = parts[0]

    if identity_expr == 'props':
      from kama_sdk.model.supplier.base.props_supplier import PropsSupplier
      identity = {'kind': PropsSupplier.__name__}
    elif identity_expr == 'ns':
      from kama_sdk.model.supplier.base.props_supplier import PropsSupplier
      identity = {'inherit': 'sdk.supplier.master_config.ns'}
    else:
      identity = {'inherit': identity_expr}

    if len(parts) == 1:
      return identity
    elif len(parts) >= 2:
      # print(f"{expr} --> #{parts} --> {parts[1:]}")
      return {**identity, 'output': " ".join(parts[1:])}
    else:
      print(f"[supplier] danger un-parsable expr {expr}")
      return {}


SER_JQ = 'jq'
SER_LEG = 'legacy'
SER_IDENTITY = 'identity'

IS_MANY_KEY = 'many'
OUTPUT_FMT_KEY = 'output'
ON_RAISE_KEY = 'on_error'
SRC_DATA_KEY = 'source'
SERIALIZER_KEY = 'serializer'



from k8kat.utils.main import units
from werkzeug.utils import cached_property

from kama_sdk.model.humanizer.quantity_humanizer import QuantityHumanizer


class CoresHumanizer(QuantityHumanizer):

  def _humanize_expr(self, raw_value: float) -> str:
    raw_value = 0 if raw_value is None else raw_value
    return units.humanize_cpu_quant(raw_value, with_unit=True)



from datetime import datetime
from typing import List, Dict, Any, Optional
from dateutil.parser import parse as parse_date
from werkzeug.utils import cached_property

from kama_sdk.core.core import utils
from kama_sdk.model.supplier.base.supplier import Supplier


class SumSupplier(Supplier):
  def _compute(self) -> float:
    values = self.source_data()
    cleanser = lambda val: float(val or 0)
    return sum(map(cleanser, values))


class ListFlattener(Supplier):
  def _compute(self) -> List:
    source = self.source_data()
    cleaner = lambda raw: raw if utils.listlike(raw) else [raw]
    return utils.flatten(list(map(cleaner, source)))


class ListFilter(Supplier):

  def predicate(self):
    from kama_sdk.model.supplier.predicate.predicate import Predicate
    return self.inflate_child(
      Predicate,
      prop='predicate',
      resolve_kod=False
    )

  @staticmethod
  def decide_sub_item(_predicate, sub_item: Any) -> bool:
    from kama_sdk.model.supplier.predicate.predicate import Predicate
    typed_predicate: Predicate = _predicate
    from kama_sdk.model.supplier.predicate import predicate
    primed_predicate: Predicate = typed_predicate.clone().patch({
      predicate.CHALLENGE_KEY: sub_item
    })
    return primed_predicate.resolve()

  def _compute(self) -> List[Any]:
    if predicate := self.predicate():
      filterer = lambda item: self.decide_sub_item(predicate, item)
      return list(filter(filterer, self.source_data()))
    else:
      return []


class ListPluck(Supplier):

  def make_flat(self) -> bool:
    return self.get_prop('flat', False)

  def resolve(self) -> List[Any]:
    included = []
    for it in self._compute():
      predicate_kod, value = it.get('include'), it.get('value')
      predicate_outcome = self.resolve_prop_value(predicate_kod)
      if utils.any2bool(predicate_outcome):
        if self.make_flat() and utils.listlike(value):
          included += value
        else:
          included.append(value)
    return included


class MergeSupplier(Supplier):

  # @cached_property
  def source_data(self) -> List[Dict]:
    dicts = super(MergeSupplier, self).source_data()
    return [d or {} for d in dicts]

  def _compute(self) -> Any:
    return utils.deep_merge(*self.source_data())


class FormattedDateSupplier(Supplier):

  def source_data(self) -> Optional[datetime]:
    value = super(FormattedDateSupplier, self).source_data()
    parse = lambda: parse_date(value)
    return utils.safely(parse) if type(value) == str else value

  @cached_property
  def output_format(self):
    return super().output_format or "%b %d at %I:%M%p %Z"

  def resolve(self) -> Any:
    source = self.source_data()
    if type(source) == datetime:
      return source.strftime(self.output_format)
    else:
      return None

class IfThenElse(Supplier):

  def on_true(self) -> Any:
    return self.resolve_prop('if_true')

  def on_false(self) -> Any:
    return self.resolve_prop('if_false')

  def _compute(self) -> Any:
    resolved_to_truthy = self.resolve_prop('source')
    return self.on_true() if resolved_to_truthy else self.on_false()



from typing import Optional, Dict

from werkzeug.utils import cached_property

from kama_sdk.model.glance.glance import Glance
from kama_sdk.model.humanizer.quantity_humanizer import QuantityHumanizer


class PercentageGlance(Glance):

  PCT_KEY = 'pct'
  FRACTION_KEY = 'fraction'
  NUMERATOR_KEY = 'numerator'
  DENOMINATOR_KEY = 'denominator'
  ABS_HUMANIZER = 'parts_humanizer'
  PCT_HUMANIZER = 'pct_humanizer'

  @property
  def view_type(self) -> str:
    return 'gauge'

  @cached_property
  def abs_humanizer(self):
    return self.inflate_child(
      QuantityHumanizer,
      prop=self.ABS_HUMANIZER,
      safely=True
    ) or QuantityHumanizer({})

  @cached_property
  def info(self) -> str:
    humanize = lambda flt: self.abs_humanizer.humanize_expr(flt)
    to_s = lambda flt: humanize(flt) if flt is not None else 'N/A'
    return f"{to_s(self.numerator)} of {to_s(self.denominator)}"

  @cached_property
  def pct(self) -> float:
    return self.get_prop(self.PCT_KEY, self.fraction * 100)

  @cached_property
  def fraction(self) -> Optional[float]:
    return self.resolve_prop(
      self.FRACTION_KEY,
      lazy_backup=lambda: self.numerator / self.denominator
    )

  @cached_property
  def numerator(self) -> Optional[float]:
    return self.get_prop(self.NUMERATOR_KEY)

  @cached_property
  def denominator(self) -> Optional[float]:
    return self.get_prop(self.DENOMINATOR_KEY)

  @cached_property
  def pct_text(self) -> str:
    return f"{str(int(self.pct))}%"

  def content_spec(self) -> Dict:
    return {
      'pct': self.pct,
      'line_one': f"{self.pct_text} of",
      'line_two': "4 Cores"
    }



from typing import Optional, Dict
from kama_sdk.core.core import updates_man, hub_api_client
from kama_sdk.core.core.types import InjectionsDesc, ErrCapture
from kama_sdk.model.action.base.action import Action
from kama_sdk.model.action.base.action_errors import FatalActionError


class FetchLatestInjectionsAction(Action):

  def perform(self) -> Optional[Dict]:
    bundle: InjectionsDesc = updates_man.latest_injection_bundle()
    raise_on_bundle_na(bundle)
    return dict(injections=bundle)


def raise_on_bundle_na(injection_bundle: Optional[InjectionsDesc]):
  if not injection_bundle:
    host = hub_api_client.backend_host()
    raise FatalActionError(ErrCapture(
      type='injection_fetch',
      name='bad_resp',
      reason=f"Negative response from Nectar API ({host})"
    ))



from flask import Blueprint, jsonify, request

from kama_sdk.core.telem.audit_sink import audit_sink
from kama_sdk.core.telem.ost import OperationState

REST_PATH = '/api/audit_items'

controller = Blueprint('auditable_events_controller', __name__)


@controller.route(f"{REST_PATH}/persist-operation-outcome")
def persist_operation_outcome():
  ost_id = request.json['ost_id']
  op_state = OperationState.find(ost_id)

  if op_state:
    audit_sink.save_op_outcome(op_state, True)



@controller.route(REST_PATH)
def index():
  query_params = None


@controller.route(f'{REST_PATH}/upload-all', methods=['POST'])
def upload_all():
  # todo start background job
  return jsonify(status='pending')



from datetime import datetime
from typing import List, Optional, Dict

from kama_sdk.core.core import utils
from kama_sdk.model.operation.step_state import StepState


class OperationState:
  def __init__(self, uuid: str, operation_id: str):
    self.uuid: str = uuid
    self.op_id: str = operation_id
    self.step_states: List[StepState] = []
    self.status = 'running'
    self.preflight_telem = None

  def notify_preflight_performed(self, telem):
    self.preflight_telem = telem

  def notify_ended(self, status):
    self.status = status

  def gen_event_telem(self) -> Dict:
    return dict(
      _id=self.uuid,
      status=self.status,
      type='operation_outcome',
      name=self.op_id,
      occurred_at=str(datetime.now())
    )

  def find_step_state(self, step) -> StepState:
    matcher = lambda ss: ss.step_sig == step.sig()
    return next(filter(matcher, self.step_states), None)

  def gen_step_state(self, step, keep=True) -> StepState:
    self.del_step_state(step)
    new_ss = StepState(step.sig(), self)
    if keep:
      self.step_states.append(new_ss)
    return new_ss

  def del_step_state(self, step):
    matcher = lambda ss: not ss.step_sig == step.sig()
    self.step_states = list(filter(matcher, self.step_states))

  def all_assigns(self) -> Dict:
    merged = {}
    for step_state in self.step_states:
      merged = {**merged, **step_state.all_assigns()}
    return merged

  def gen_sub_map(self):
    assigns = self.all_assigns()
    keyed = utils.dict2keyed(assigns)
    return {f"$var/{k}": v for k, v in keyed}

  @classmethod
  def find(cls, ost_id: Optional[str]) -> 'OperationState':
    """
    Finds the OperationState with the marching osr id.
    :param ost_id: id to match by.
    :return: OperationState or None
    """
    matcher = lambda op_state: op_state.uuid == ost_id
    return next(filter(matcher, operation_states), None)

  @classmethod
  def gen(cls, operation_id: str) -> str:
    uuid = utils.rand_str(string_len=10)
    operation_states.append(OperationState(uuid, operation_id))
    return uuid

  @classmethod
  def delete_if_exists(cls, ost_id: Optional[str]) -> Optional['OperationState']:
    """
    Deletes the OperationState if one with the given osr id exists.
    :param ost_id: id used to locate the OperationState.
    :return: deleted instance of the OperationState if found, else None.
    """
    tuples = enumerate(operation_states)
    index = next((i for i, ops in tuples if ops.uuid == ost_id), None)
    return operation_states.pop(index) if index else None

  @classmethod
  def clear_list(cls):
    operation_states.clear()

  @classmethod
  def prune(cls):
    while cls.find(None):
      cls.delete_if_exists(None)


operation_states: List[OperationState] = []



from typing import Optional, Dict

from k8kat.res.base.kat_res import KatRes
from werkzeug.utils import cached_property

from kama_sdk.core.core.config_man import config_man
from kama_sdk.core.core.types import ErrCapture


class ErrCtx:
  def __init__(self, errdict: ErrCapture):
    self._errdict: Dict = errdict

  @cached_property
  def selectable_properties(self) -> Dict:
    all_keys = self._errdict.keys()
    avoid_keys = ['uuid', 'resource']
    good_keys = list(set(all_keys) - set(avoid_keys))
    base_level = {k: self._errdict[k] for k in good_keys}
    extras_level = self._errdict.get('extras', {})
    return {**base_level, **extras_level}

  @cached_property
  def event_type(self):
    return self._errdict.get('event_type')

  @cached_property
  def kubernetes_resource(self) -> Optional[KatRes]:
    res_dict = self._errdict.get('resource')
    return self.extract_kat_res(res_dict) if res_dict else None

  @cached_property
  def resource_dict(self) -> Dict:
    return self._errdict.get('resource')

  @staticmethod
  def extract_kat_res(res_desc: Dict) -> Optional[KatRes]:
    kind, name = res_desc.get('kind'), res_desc.get('name')
    if kind and name:
      kat_model = KatRes.class_for(kind)
      return kat_model.find(name, config_man.ns()) if kat_model else None
    return None
